import logging

LOGGER = logging.getLogger(__name__)

CALL_INS = ["call", "calli", "callvirt"]
END_INS = ["ret"]

class FunctionAnalysisState(object):

    def __init__(self, start_addr, code_start_addr, disassembly):
        self.start_addr = start_addr
        self.code_start_addr = code_start_addr
        self.disassembly = disassembly
        self.block_queue = [start_addr]
        self.current_block = []
        self.blocks = []
        self.num_blocks_analyzed = 0
        self.instructions = []
        self.instruction_start_bytes = set([])
        self.processed_blocks = set([])
        self.processed_bytes = set([])
        self.jump_targets = set([])
        self.call_register_ins = []
        self.block_start = 0xFFFFFFFF
        self.data_bytes = set([])
        self.data_refs = set([])
        self.code_refs = set([])
        self.code_refs_from = {}
        self.code_refs_to = {}
        self.prev_opcode = ""
        self.suspicious_ins_count = 0
        self.is_jmp = False
        self.is_next_instruction_reachable = True
        self.is_block_ending_instruction = False
        self.is_sanely_ending = False
        self.has_collision = False
        self.colliding_addresses = set()
        # set a flag that this tailcall has already been resolved so it does not have to be reanalyzed several times
        self.is_tailcall_function = False
        self.is_leaf_function = True
        self.is_recursive = False
        self.is_thunk_call = False
        self.label = ""

    def addInstruction(self, i_address, i_size, i_mnemonic, i_op_str, i_bytes):
        ins = (i_address, i_size, i_mnemonic, i_op_str, i_bytes)
        self.instructions.append(ins)
        self.instruction_start_bytes.add(ins[0])
        for byte in range(i_size):
            self.processed_bytes.add(i_address + byte)
        if self.is_next_instruction_reachable:
            self.addCodeRef(i_address, i_address + i_size, self.is_jmp)
        self.is_jmp = False

    def addCodeRef(self, addr_from, addr_to, by_jump=False):
        self.code_refs.update([(addr_from, addr_to)])
        refs_from = self.code_refs_from.get(addr_from, set([]))
        refs_from.update([addr_to])
        self.code_refs_from[addr_from] = refs_from
        refs_to = self.code_refs_to.get(addr_to, set([]))
        refs_to.update([addr_from])
        self.code_refs_to[addr_to] = refs_to
        if by_jump:
            self.is_jmp = True
            self.jump_targets.update([addr_to])

    def removeCodeRef(self, addr_from, addr_to):
        if (addr_from, addr_to) in self.code_refs:
            self.code_refs.remove((addr_from, addr_to))
        if addr_from in self.code_refs_from and addr_to in self.code_refs_from[addr_from]:
            self.code_refs_from[addr_from].remove(addr_to)
        if addr_to in self.code_refs_to and addr_from in self.code_refs_to[addr_to]:
            self.code_refs_to[addr_to].remove(addr_from)
        if addr_to in self.jump_targets:
            self.jump_targets.remove(addr_to)

    def addDataRef(self, addr_from, addr_to, size=1):
        self.data_refs.update([(addr_from, addr_to)])
        for i in range(size):
            self.data_bytes.update([addr_to + i])

    def finalizeAnalysis(self, as_gap=False):
        fn_min = min([ins[0] for ins in self.instructions])
        fn_max = max([ins[0] + ins[1] for ins in self.instructions])

        self.disassembly.function_symbols[self.start_addr] = self.label
        self.disassembly.function_borders[self.start_addr] = (fn_min, fn_max)
        for ins in self.instructions:
            self.disassembly.instructions[ins[0]] = (ins[2], ins[1])
            for offset in range(ins[1]):
                self.disassembly.code_map[ins[0] + offset] = ins[0]
                self.disassembly.ins2fn[ins[0] + offset] = self.start_addr
        self.disassembly.data_map.update(self.data_bytes)
        self.disassembly.functions[self.start_addr] = self.getBlocks()
        for cref in self.code_refs:
            self.disassembly.addCodeRefs(cref[0], cref[1])
        for dref in self.data_refs:
            self.disassembly.addDataRefs(dref[0], dref[1])
        if self.is_recursive:
            self.disassembly.recursive_functions.add(self.start_addr)
        if self.is_leaf_function:
            self.disassembly.leaf_functions.add(self.start_addr)
        if self.is_thunk_call:
            self.disassembly.thunk_functions.add(self.start_addr)
        return True

    def getBlocks(self):
        """
        block derivation strategy:
        walk over all potential block starts, which are the start_addr + all "jump" targets (i.e. CFG redirection targets)
        then, for consecutive instructions, break if
        * they have more than 1 outgoing edge
        * the following instruction has more than 1 incoming edge
        """
        if self.blocks:
            return self.blocks
        self.instructions.sort()
        ins = {i[0]:ind for ind, i in enumerate(self.instructions)}
        potential_starts = set([self.code_start_addr])
        potential_starts.update(list(self.jump_targets))
        blocks = []
        for start in sorted(potential_starts):
            if not start in ins:
                continue
            block = []
            for i in range(ins[start], len(self.instructions)):
                current = self.instructions[i]
                block.append(current)
                # if one code reference is to another address than the next
                if current[0] in self.code_refs_from:
                    if not current[2] in CALL_INS and not i == len(self.instructions) - 1:
                        if any([r != self.instructions[i+1][0] for r in self.code_refs_from[current[0]]]):
                            break
                    # if we can reach a colliding address from here, the block is broken and should end.
                    reachable_collisions = self.code_refs_from[current[0]].intersection(self.colliding_addresses)
                    next_addr = current[0] + current[1]
                    is_next_addr = next_addr in reachable_collisions
                    if reachable_collisions and is_next_addr:
                        # we should remove the from/to code references for this collision as there should be no non CFG instruction references between instructions of different functions
                        self.removeCodeRef(current[0], next_addr)
                        break
                if not i == len(self.instructions) - 1 and self.instructions[i+1][0] in self.code_refs_to:
                    if len(self.code_refs_to[self.instructions[i+1][0]]) > 1 or self.instructions[i+1][0] in potential_starts:
                        break
                if current[2] in END_INS:
                    break
            if block:
                blocks.append(block)
        self.blocks = blocks
        return self.blocks

    def isProcessed(self, addr):
        return addr in self.processed_bytes

    def isProcessedFunction(self):
        return self.start_addr in self.disassembly.code_map

    def isNextInstructionReachable(self):
        return self.is_next_instruction_reachable

    def setNextInstructionReachable(self, is_reachable):
        self.is_next_instruction_reachable = is_reachable

    def __str__(self):
        result = "0x{:x} | current: 0x{:x} | blocks: {} | queue: {} | processed: {} | crefs: {} | drefs: {} | suspicious: {} | ending: {}".format(
            self.start_addr,
            self.block_start,
            len(self.getBlocks()),
            ",".join(["0x%x" % b for b in sorted(self.block_queue)]),
            ",".join(["0x%x" % b for b in sorted(list(self.processed_blocks))]),
            len(self.code_refs),
            len(self.data_refs),
            self.suspicious_ins_count,
            self.is_sanely_ending
        )
        return result



#!/usr/bin/python

import logging
import os
import unittest

from smda.common.SmdaReport import SmdaReport
from smda.common.SmdaFunction import SmdaFunction
from smda.common.SmdaInstruction import SmdaInstruction
from smda.intel.IntelInstructionEscaper import IntelInstructionEscaper
from smda.cil.CilInstructionEscaper import CilInstructionEscaper

from .context import config

LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
logging.disable(logging.CRITICAL)


class DisassemblyTestSuite(unittest.TestCase):
    """Run a full example on a memory dump"""

    def testInstructionEscaping(self):
        test_data = [
            {"ins": (0, "55", "push", "ebp"), "mnemonic_group": "S", "escaped_operands": "REG"},
            {"ins": (1, "8365fc00", "and", "dword ptr [ebp - 4], 0"), "mnemonic_group": "A", "escaped_operands": "PTR, CONST"},
            {"ins": (2, "f30f1efa", "endbr64", ""), "mnemonic_group": "C", "escaped_operands": ""},
            {"ins": (3, "c58e5ad3", "vcvtss2sd", "xmm2, xmm14, xmm3"), "mnemonic_group": "X", "escaped_operands": "XREG, XREG, XREG"},
        ]
        for data in test_data:
            smda_ins = SmdaInstruction(data["ins"])
            self.assertEqual(smda_ins.getMnemonicGroup(IntelInstructionEscaper), data["mnemonic_group"])
            self.assertEqual(smda_ins.getEscapedOperands(IntelInstructionEscaper), data["escaped_operands"])

    def testIntelInstructionWildcarding(self):
        test_data = [
            # simple mov with IMM outside of address space
            {"ins": (0, "b803400080", "mov", "eax, 0x80004003"), "lower": 0x63300000, "upper": 0x63400000, "expected_bin": "b803400080", "bitness": 32, "expected_opc": "b8????????"},
            # simple mov with IMM within address space
            {"ins": (0, "ba2c893863", "mov", "edx, 0x6338892c"), "lower": 0x63300000, "upper": 0x63400000, "expected_bin": "ba????????", "bitness": 32, "expected_opc": "ba????????"},
            # mov with with address calc within address space
            {"ins": (0, "0fb681808f3b63", "mov", "eax, byte ptr [ecx + 0x633b8f80]"), "lower": 0x63300000, "upper": 0x63400000, "expected_bin": "0fb681????????", "bitness": 32, "expected_opc": "0fb6??????????"},
            # jump table calculation
            {"ins": (0, "ff2485788f3b63", "jmp", "dword ptr [eax*4 + 0x633b8f78]"), "lower": 0x63300000, "upper": 0x63400000, "expected_bin": "ff2485????????", "bitness": 32, "expected_opc": "ff????????????"},
            # should only wildcard last part as escaper doesn't know address space
            {"ins": (0, "c705ac974a00ac974a00", "mov", "dword ptr [0x4a97ac], 0x4a97ac"), "lower": None, "upper": None, "expected_bin": "c705ac974a00????????", "bitness": 32, "expected_opc": "c7??????????????????"},
            # should escape both operands
            {"ins": (0, "c705ac974a00ac974a00", "mov", "dword ptr [0x4a97ac], 0x4a97ac"), "lower": 0x400000, "upper": 0x4f0000, "expected_bin": "c705????????????????", "bitness": 32, "expected_opc": "c7??????????????????"},
            # should escape from the right side and only blank out one, despite finding two matches for the pattern
            {"ins": (0, "010505050505", "add", "dword ptr [0x5050505], eax"), "lower": 0x400000, "upper": 0x4f0000, "expected_bin": "0105????????", "bitness": 32, "expected_opc": "01??????????"},
            # should escape from the right side and only blank out one, despite finding two matches for the pattern
            {"ins": (0, "0f101515151515", "movups", "xmm2, xmmword ptr [0x15151515]"), "lower": 0x400000, "upper": 0x4f0000, "expected_bin": "0f1015????????", "bitness": 32, "expected_opc": "0f10??????????"},
            # should ignore prefixes while wildcarding
            {"ins": (0, "666666660f008000224000", "sldt", "word ptr [rax + 0x402200]"), "lower": 0x400000, "upper": 0x4f0000, "expected_bin": "666666660f0080????????", "bitness": 32, "expected_opc": "666666660f00??????????"},
            # should ignore prefixes and REX while wildcarding
            {"ins": (0, "66666666480f008000224000", "sldt", "word ptr [rax + 0x402200]"), "lower": 0x400000, "upper": 0x4f0000, "expected_bin": "66666666480f0080????????", "bitness": 64, "expected_opc": "66666666480f00??????????"},
        ]
        for data in test_data:
            smda_report = SmdaReport()
            smda_report.bitness = data["bitness"]
            smda_function = SmdaFunction(smda_report=smda_report)
            smda_ins = SmdaInstruction(data["ins"], smda_function=smda_function)
            self.assertEqual(smda_ins.getEscapedBinary(IntelInstructionEscaper, lower_addr=data["lower"], upper_addr=data["upper"]), data["expected_bin"])
            self.assertEqual(smda_ins.getEscapedToOpcodeOnly(IntelInstructionEscaper), data["expected_opc"])

    def testCilInstructionWildcarding(self):
        test_data = [
            # call MemberRef
            {"ins": (0, "280a000006", "call", "SomeFunc"), "expected_bin": "28??????06", "expected_bin_intraprocedural": "28??????06", "expected_opc": "28????????", "bitness": 32},
            {"ins": (0, "6fbb00000a", "callvirt", "SomeFunc"), "expected_bin": "6f??????0a", "expected_bin_intraprocedural": "6f??????0a", "expected_opc": "6f????????", "bitness": 32},
            {"ins": (0, "2d3a", "brtrue.s", "0x5994"), "expected_bin": "2d3a", "expected_bin_intraprocedural": "2d??", "expected_opc": "2d??", "bitness": 32},
            {"ins": (0, "450300000002000000060000000a000000", "switch", "[(0D50), (0D54), (0D58)]"), "expected_bin": "450300000002000000060000000a000000", "expected_bin_intraprocedural": "45????????????????????????????????", "expected_opc": "45????????????????????????????????", "bitness": 32},
            {"ins": (0, "20c48efb0e", "ldc.i4", "0xefb8ec4"), "expected_bin": "20c48efb0e", "expected_bin_intraprocedural": "20c48efb0e", "expected_opc": "20????????", "bitness": 32},
        ]
        for data in test_data:
            smda_report = SmdaReport()
            smda_report.bitness = data["bitness"]
            smda_function = SmdaFunction(smda_report=smda_report)
            smda_ins = SmdaInstruction(data["ins"], smda_function=smda_function)
            self.assertEqual(CilInstructionEscaper.escapeToOpcodeOnly(smda_ins), data["expected_opc"])
            self.assertEqual(CilInstructionEscaper.escapeBinary(smda_ins), data["expected_bin"])
            self.assertEqual(CilInstructionEscaper.escapeBinary(smda_ins, escape_intraprocedural_jumps=True), data["expected_bin_intraprocedural"])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python

import logging
import os
import unittest

from smda.common.Tarjan import Tarjan

from .context import config

LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
logging.disable(logging.CRITICAL)


class TarjanTestSuite(unittest.TestCase):
    """Provoke recursion"""

    def testInstructionEscaping(self):
        test_data = {i: [] for i in range(1000)}
        for i in range(1, 1000):
            for j in range(i + 1, 1000, 1):
                test_data[i].append(j)
        test_data[1000] = []

        tarjan = Tarjan(test_data)
        tarjan.calculateScc()
        sccs = tarjan.getResult()
        self.assertEqual(1001, len(sccs))


if __name__ == '__main__':
    unittest.main()



class BracketQueue(object):
    """
    This queue is tailored based on our research rsults regarding function entry point identification
    """
    def __init__(self, candidates=None, initial_brackets=None):
        self.update_count = 0
        self.update_shift_count = 0
        self.brackets = {
            0: {},
            1: {},
            2: {}
        }
        if candidates is not None:
            for candidate in candidates:
                self.add(candidate)
            self.ensure_order()
        elif initial_brackets is not None:
            self.brackets = initial_brackets
            self.ensure_order()
        
    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if all(len(self.brackets[i]) == 0 for i in range(3)):
            raise StopIteration
        for bracket_index in range(2, -1, -1):
            if self.brackets[bracket_index]:
                offset, candidate = self.brackets[bracket_index].popitem()
                return candidate

    def add(self, candidate):
        bracket_index = min(2, len(candidate.call_ref_sources))
        self.brackets[bracket_index][candidate.addr] = candidate

    def update(self, target_candidate=None):
        if target_candidate:
            updated_bracket_index = min(2, len(target_candidate.call_ref_sources))
            # check if the element is still in the same bracket, otherwise shift to next bracket
            self.update_count += 1
            for bracket_index in range(2, -1, -1):
                if target_candidate.addr in self.brackets[bracket_index] and bracket_index != updated_bracket_index:
                    self.update_shift_count += 1
                    self.brackets[bracket_index].pop(target_candidate.addr)
                    self.brackets[updated_bracket_index][target_candidate.addr] = target_candidate
                    break

    def ensure_order(self):
        for bracket_index in range(2, -1, -1):
            if self.brackets[bracket_index]:
                self.brackets[bracket_index] = {offset: candidate for offset, candidate in sorted(self.brackets[bracket_index].items(), key=lambda x: x[1].getScore())}

    def __str__(self):
        return f"BracketQueue | 2: {len(self.brackets[2])} candidates, 1: {len(self.brackets[1])} candidates, 0: {len(self.brackets[0])} candidates,"



#!/usr/bin/python

import logging
import os
import unittest

from smda.utility.FileLoader import FileLoader
from smda.common.BinaryInfo import BinaryInfo
from smda.Disassembler import Disassembler
from smda.common.SmdaReport import SmdaReport
from smda.common.SmdaFunction import SmdaFunction
from .context import config

LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
logging.disable(logging.CRITICAL)


class SmdaIntegrationTestSuite(unittest.TestCase):
    """Run a full example on a memory dump"""

    @classmethod
    def setUpClass(cls):
        super(SmdaIntegrationTestSuite, cls).setUpClass()
        config.WITH_STRINGS = True
        disasm = Disassembler(config)
        # load encrypted Asprox
        with open(os.path.join(config.PROJECT_ROOT, "tests", "asprox_0x008D0000_xored"), "rb") as f_binary:
            binary = f_binary.read()
        decrypted_asprox = bytearray()
        for index, byte in enumerate(binary):
            if isinstance(byte, str):
                byte = ord(byte)
            decrypted_asprox.append(byte ^ (index % 256))
        cls.asprox_binary = decrypted_asprox
        cls.asprox_disassembly = disasm.disassembleBuffer(bytes(decrypted_asprox), 0x8D0000)
        # load encrypted Cutwail
        with open(os.path.join(config.PROJECT_ROOT, "tests", "cutwail_xored"), "rb") as f_binary:
            binary = f_binary.read()
        decrypted_cutwail = bytearray()
        for index, byte in enumerate(binary):
            if isinstance(byte, str):
                byte = ord(byte)
            decrypted_cutwail.append(byte ^ (index % 256))
        cls.cutwail_binary = bytes(decrypted_cutwail)
        # run FileLoader and disassemble as file
        loader = FileLoader("/", map_file=True)
        loader._loadFile(cls.cutwail_binary)
        file_content = loader.getData()
        binary_info = BinaryInfo(file_content)
        binary_info.raw_data = loader.getRawData()
        binary_info.file_path = ""
        binary_info.base_addr = loader.getBaseAddress()
        binary_info.bitness = loader.getBitness()
        binary_info.code_areas = loader.getCodeAreas()
        binary_info.oep = binary_info.getOep()
        cls.cutwail_binary_info = binary_info
        cls.cutwail_disassembly = disasm._disassemble(binary_info)
        cls.cutwail_unmapped_disassembly = disasm.disassembleUnmappedBuffer(cls.cutwail_binary)

    def testAsproxDisassemblyCoverage(self):
        assert len([fn for fn in self.asprox_disassembly.getFunctions()]) == 105

    def testOep(self):
        # PE header from buffers are not parsed, so we don't get header infos
        assert self.asprox_disassembly.oep == None
        # PE headers are parsed for regularly processed files (PE+ELF)
        assert self.cutwail_unmapped_disassembly.oep == 0x1730

    def testCodeXrefCreation(self):
        example_function = self.asprox_disassembly.getFunction(0x008d8292)
        # should be initialized on demand only
        assert example_function.code_inrefs == None
        # example function has inrefs and outrefs
        inrefs = [code_inref for code_inref in example_function.getCodeInrefs()]
        assert len(inrefs) == 1
        for xref in example_function.getCodeInrefs():
            print(xref.from_function, xref.from_instruction, xref.to_function, xref.to_instruction)
        outrefs = [code_outref for code_outref in example_function.getCodeOutrefs()]
        assert len(outrefs) == 10

    def testAsproxStringRefs(self):
        function_with_strings = self.asprox_disassembly.getFunction(0x008d2850)
        assert function_with_strings.stringrefs[9251000] == "Software"
        marshalled = function_with_strings.toDict()
        unmarshalled = SmdaFunction.fromDict(marshalled)
        assert unmarshalled.stringrefs[9251000] == "Software"

    def testAsproxApiCoverage(self):
        num_api_ref_srcs = 0
        api_ref_dsts = set()
        for fn in self.asprox_disassembly.getFunctions():
            num_api_ref_srcs += len(fn.apirefs)
            api_ref_dsts.update(fn.apirefs.values())
        assert num_api_ref_srcs == 546
        assert len(api_ref_dsts) == 95

    def testAsproxMarshalling(self):
        report_as_dict = self.asprox_disassembly.toDict()
        assert report_as_dict["status"] == "ok"
        assert report_as_dict["base_addr"] == 0x8D0000
        assert report_as_dict["statistics"]["num_instructions"] == 15706
        assert report_as_dict["sha256"] == "db8a133fed1b706608a4492079b702ded6b70369a980d2b5ae355a6adc78ef00"
        reimported_report = SmdaReport.fromDict(report_as_dict)

    def testCutwailMarshalling(self):
        report_as_dict = self.cutwail_disassembly.toDict()
        assert report_as_dict["status"] == "ok"
        assert report_as_dict["base_addr"] == 0x4000000
        assert report_as_dict["statistics"]["num_instructions"] == 1611
        assert report_as_dict["sha256"] == "a348a0ddfab135d152b684d561a3215ab6c472570facd3d75aa2c7ee845a8e2b"
        # compare our manual file loading with unmapped buffer
        assert self.cutwail_disassembly.num_instructions == self.cutwail_unmapped_disassembly.num_instructions
        reimported_report = SmdaReport.fromDict(report_as_dict)

    def testBlockLocator(self):
        # test with a function start
        found_function = self.asprox_disassembly.findFunctionByContainedAddress(0x008d8292)
        found_block = self.asprox_disassembly.findBlockByContainedAddress(0x008d8292)
        assert found_function.offset == 0x008d8292
        assert found_block.offset == 0x008d8292
        # test with an instruction in a block a bit deeper in the function
        found_function = self.asprox_disassembly.findFunctionByContainedAddress(0x008d82a6)
        found_block = self.asprox_disassembly.findBlockByContainedAddress(0x008d82a6)
        assert found_function.offset == 0x008d8292
        assert found_block.offset == 0x008d82a4
        # test with an offset that is not start of an instruction
        found_function = self.asprox_disassembly.findFunctionByContainedAddress(0x008d82a7)
        found_block = self.asprox_disassembly.findBlockByContainedAddress(0x008d82a7)
        assert found_function.offset == 0x008d8292
        assert found_block.offset == 0x008d82a4
        # test with offsets beyond image base and binary size
        found_function = self.asprox_disassembly.findFunctionByContainedAddress(0x100)
        found_block = self.asprox_disassembly.findBlockByContainedAddress(0x100)
        assert found_function is None
        assert found_block is None
        found_function = self.asprox_disassembly.findFunctionByContainedAddress(0xFFFFFF00)
        found_block = self.asprox_disassembly.findBlockByContainedAddress(0xFFFFFF00)
        assert found_function is None
        assert found_block is None


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python

import logging
from .AbstractLabelProvider import AbstractLabelProvider

LOGGER = logging.getLogger(__name__)

try:
    import lief
    lief.logging.disable()
except:
    lief = None
    LOGGER.warning("3rd party library LIEF not installed - won't be able to extract symbols for ELF files where available.")



class PeSymbolProvider(AbstractLabelProvider):
    """ Minimal resolver for PE symbols """

    def __init__(self, config):
        self._config = config
        #addr:func_name
        self._func_symbols = {}

    def isSymbolProvider(self):
        return True

    def _parseOep(self, lief_result):
        if lief_result:
            self._func_symbols[lief_result.entrypoint] = "original_entry_point"

    def update(self, binary_info):
        #works both for PE and ELF
        self._func_symbols = {}
        if not binary_info.file_path:
            return
        data = ""
        with open(binary_info.file_path, "rb") as fin:
            data = fin.read(16)
        if data[:2] != b"MZ" or lief is None:
            return
        lief_binary = lief.parse(binary_info.file_path)
        if lief_binary is not None:
            self._parseOep(lief_binary)
            self._parseExports(lief_binary)
            self._parseSymbols(lief_binary)

    def _parseExports(self, binary):
        for function in binary.exported_functions:
            function_name = ""
            try:
                # here may occur a LIEF exception that we want to skip ->
                # UnicodeDecodeError: 'utf-32-le' codec can't decode bytes in position 0-3: code point not in range(0x110000)
                function_name = function.name
            except:
                pass
            if function_name and all(c in range(0x20, 0x7f) for c in function_name):
                self._func_symbols[binary.imagebase + function.address] = function_name

    def _parseSymbols(self, lief_binary):
        # find VA of first code section
        code_base_address = None
        for section in lief_binary.sections:
            if section.characteristics & 0x20000000:
                code_base_address = lief_binary.imagebase + section.virtual_address
                break
        if code_base_address is None:
            return
        for symbol in lief_binary.symbols:
            if hasattr(symbol.complex_type, "name") and symbol.complex_type.name == "FUNCTION":
                function_name = ""
                try:
                    # here may occur a LIEF exception that we want to skip ->
                    # UnicodeDecodeError: 'utf-32-le' codec can't decode bytes in position 0-3: code point not in range(0x110000)
                    function_name = symbol.name
                except:
                    pass
                if function_name and all(ord(c) in range(0x20, 0x7f) for c in function_name):
                    # for some reason, we need to add the section_offset of .text here
                    function_offset = code_base_address + symbol.value
                    if function_offset not in self._func_symbols:
                        self._func_symbols[function_offset] = function_name

    def getSymbol(self, address):
        return self._func_symbols.get(address, "")

    def getFunctionSymbols(self):
        return self._func_symbols



#!/usr/bin/python

import logging
import os
import lief
import unittest

from smda.utility.FileLoader import FileLoader
from smda.common.BinaryInfo import BinaryInfo
from smda.Disassembler import Disassembler
from smda.common.SmdaReport import SmdaReport
from smda.common.SmdaFunction import SmdaFunction
from .context import config

LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
logging.disable(logging.CRITICAL)


class SmdaIntegrationTestSuite(unittest.TestCase):
    """Run a full example on a memory dump"""

    @classmethod
    def setUpClass(cls):
        super(SmdaIntegrationTestSuite, cls).setUpClass()

    def testPeParsingWithCutwail(self):
        disasm = Disassembler(config, backend="intel")
        # load encrypted malicious win.cutwail
        with open(os.path.join(config.PROJECT_ROOT, "tests", "cutwail_xored"), "rb") as f_binary:
            binary = f_binary.read()
        decrypted_cutwail = bytearray()
        for index, byte in enumerate(binary):
            if isinstance(byte, str):
                byte = ord(byte)
            decrypted_cutwail.append(byte ^ (index % 256))
        cutwail_binary = bytes(decrypted_cutwail)
        # run FileLoader and disassemble as file
        loader = FileLoader("/", map_file=True)
        loader._loadFile(cutwail_binary)
        file_content = loader.getData()
        binary_info = BinaryInfo(file_content)
        binary_info.raw_data = loader.getRawData()
        binary_info.file_path = ""
        binary_info.base_addr = loader.getBaseAddress()
        binary_info.bitness = loader.getBitness()
        binary_info.code_areas = loader.getCodeAreas()
        binary_info.oep = binary_info.getOep()
        cutwail_binary_info = binary_info
        # parse bytes of 0x400 truncated PE header
        pe_header = lief.parse(binary_info.getHeaderBytes())
        assert pe_header.dos_header.magic == 0x5A4D
        assert pe_header.header.machine == 0x14C
        cutwail_disassembly = disasm._disassemble(binary_info)
        cutwail_unmapped_disassembly = disasm.disassembleUnmappedBuffer(cutwail_binary)
        assert cutwail_unmapped_disassembly.num_functions == 33
        # TODO test label extraction for PE, add another binary for testing

    def testElfParsingWithBashlite(self):
        disasm = Disassembler(config, backend="intel")
        # load encrypted benign /bin/cat
        with open(os.path.join(config.PROJECT_ROOT, "tests", "bashlite_xored"), "rb") as f_binary:
            binary = f_binary.read()
        decrypted_bashlite = bytearray()
        for index, byte in enumerate(binary):
            if isinstance(byte, str):
                byte = ord(byte)
            decrypted_bashlite.append(byte ^ (index % 256))
        bashlite_binary = bytes(decrypted_bashlite)
        # run FileLoader and disassemble as file
        loader = FileLoader("/", map_file=True)
        loader._loadFile(bashlite_binary)
        file_content = loader.getData()
        binary_info = BinaryInfo(file_content)
        binary_info.raw_data = loader.getRawData()
        binary_info.file_path = ""
        binary_info.base_addr = loader.getBaseAddress()
        binary_info.bitness = loader.getBitness()
        binary_info.code_areas = loader.getCodeAreas()
        binary_info.oep = binary_info.getOep()
        bashlite_binary_info = binary_info
        bashlite_disassembly = disasm._disassemble(binary_info)
        bashlite_unmapped_disassembly = disasm.disassembleUnmappedBuffer(bashlite_binary)
        assert bashlite_unmapped_disassembly.num_functions == 177
        assert len([f.function_name for f in bashlite_unmapped_disassembly.getFunctions() if f.function_name]) == 174

    def testDotnetParsingWithNjRAT(self):
        disasm = Disassembler(config, backend="cil")
        # load encrypted malicious win.cutwail
        with open(os.path.join(config.PROJECT_ROOT, "tests", "njrat_xored"), "rb") as f_binary:
            binary = f_binary.read()
        decrypted_njrat = bytearray()
        for index, byte in enumerate(binary):
            if isinstance(byte, str):
                byte = ord(byte)
            decrypted_njrat.append(byte ^ (index % 256))
        njrat_binary = bytes(decrypted_njrat)
        # run FileLoader and disassemble as file
        njrat_unmapped_disassembly = disasm.disassembleUnmappedBuffer(njrat_binary)
        assert njrat_unmapped_disassembly.num_functions == 64
        assert len([f.function_name for f in njrat_unmapped_disassembly.getFunctions() if f.function_name]) == 64

    def testMacOsParsingWithKomplex(self):
        disasm = Disassembler(config, backend="intel")
        # load encrypted malicious osx.komplex
        with open(os.path.join(config.PROJECT_ROOT, "tests", "komplex_xored"), "rb") as f_binary:
            binary = f_binary.read()
        decrypted_komplex = bytearray()
        for index, byte in enumerate(binary):
            if isinstance(byte, str):
                byte = ord(byte)
            decrypted_komplex.append(byte ^ (index % 256))
        komplex_binary = bytes(decrypted_komplex)
        # run FileLoader and disassemble as file
        loader = FileLoader("/", map_file=True)
        loader._loadFile(komplex_binary)
        file_content = loader.getData()
        binary_info = BinaryInfo(file_content)
        binary_info.raw_data = loader.getRawData()
        binary_info.file_path = ""
        binary_info.base_addr = loader.getBaseAddress()
        binary_info.bitness = loader.getBitness()
        binary_info.code_areas = loader.getCodeAreas()
        binary_info.oep = binary_info.getOep()
        komplex_binary_info = binary_info
        komplex_disassembly = disasm._disassemble(binary_info)
        komplex_unmapped_disassembly = disasm.disassembleUnmappedBuffer(komplex_binary)
        komplex_unmapped_disassembly.num_functions == 208


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python

import datetime
import logging
import re
from typing import TYPE_CHECKING, Any, Union, Optional

if TYPE_CHECKING:
    from dnfile import dnPE
    from dnfile.mdtable import MethodDefRow
import dnfile
from dnfile.enums import MetadataTables
from dncil.cil.body import CilMethodBody
from dncil.cil.error import MethodBodyFormatError
from dncil.clr.token import Token, StringToken, InvalidToken
from dncil.cil.body.reader import CilMethodBodyReaderBase

from smda.DisassemblyResult import DisassemblyResult
from smda.common.BinaryInfo import BinaryInfo
from smda.common.labelprovider.CilSymbolProvider import CilSymbolProvider
from .FunctionAnalysisState import FunctionAnalysisState

LOGGER = logging.getLogger(__name__)
DOTNET_META_TABLES_BY_INDEX = {table.value: table.name for table in MetadataTables}


def read_dotnet_user_string(pe, token: StringToken) -> Union[str, InvalidToken]:
    """read user string from #US stream"""
    try:
        user_string: Optional[dnfile.stream.UserString] = pe.net.user_strings.get(token.rid)
    except UnicodeDecodeError as e:
        return InvalidToken(token.value)

    if user_string is None or (isinstance(user_string, bytes) or user_string.value is None):
        return InvalidToken(token.value)

    return user_string.value


def resolve_token(pe, token: Token) -> Any:
    """ """
    if isinstance(token, StringToken):
        return read_dotnet_user_string(pe, token)

    table_name: str = DOTNET_META_TABLES_BY_INDEX.get(token.table, "")
    if not table_name:
        # table_index is not valid
        return InvalidToken(token.value)

    table: Any = getattr(pe.net.mdtables, table_name, None)
    if table is None:
        # table index is valid but table is not present
        return InvalidToken(token.value)

    try:
        return table.rows[token.rid - 1]
    except IndexError:
        # table index is valid but row index is not valid
        return InvalidToken(token.value)


def format_operand(pe, operand: Any) -> str:
    """ """
    if isinstance(operand, Token):
        operand = resolve_token(pe, operand)
    if isinstance(operand, str):
        return f'"{operand}"'
    elif isinstance(operand, int):
        return hex(operand)
    elif isinstance(operand, list):
        return f"[{', '.join(['({:04X})'.format(x) for x in operand])}]"
    elif isinstance(operand, dnfile.mdtable.MemberRefRow):
        if isinstance(operand.Class.row, (dnfile.mdtable.TypeRefRow,)):
            return f"{str(operand.Class.row.TypeNamespace)}.{operand.Class.row.TypeName}::{operand.Name}"
        else:
            return f"{operand.Name}"
    elif isinstance(operand, dnfile.mdtable.MethodSpecRow):
        operand = operand.Method.row
        if isinstance(operand, (dnfile.mdtable.TypeRefRow,)):
            return f"{str(operand.TypeNamespace)}.{operand.TypeName}::{operand.Name}"
        else:
            return f"{operand.Name}"
    elif isinstance(operand, dnfile.mdtable.TypeRefRow):
        return f"{str(operand.TypeNamespace)}.{operand.TypeName}"
    elif isinstance(operand, (dnfile.mdtable.FieldRow, dnfile.mdtable.MethodDefRow)):
        return f"{operand.Name}"
    elif operand is None:
        return ""

    return str(operand)


class DnfileMethodBodyReader(CilMethodBodyReaderBase):
    def __init__(self, pe, row):
        """ """
        self.pe: dnPE = pe
        self.offset: int = self.pe.get_offset_from_rva(row.Rva)

    def read(self, n: int) -> bytes:
        """ """
        data: bytes = self.pe.get_data(self.pe.get_rva_from_offset(self.offset), n)
        self.offset += n
        return data

    def tell(self) -> int:
        """ """
        return self.offset

    def seek(self, offset: int) -> int:
        """ """
        self.offset = offset
        return self.offset


class CilDisassembler(object):

    def __init__(self, config):
        self.config = config
        self._tfidf = None
        self.binary_info = None
        self.label_providers = []
        self.cil_label_provider = CilSymbolProvider(self.config)
        self._addLabelProviders()
        self.disassembly = DisassemblyResult()
        self.disassembly.smda_version = config.VERSION

    def addPdbFile(self, binary_info, pdb_path):
        return

    def _addLabelProviders(self):
        self.label_providers.append(self.cil_label_provider)

    def _updateLabelProviders(self, binary_info):
        for provider in self.label_providers:
            provider.update(binary_info)

    def resolveSymbol(self, address):
        for provider in self.label_providers:
            if not provider.isSymbolProvider():
                continue
            result = provider.getSymbol(address)
            if result:
                return result
        return ""

    def _updateApiInformation(self, from_addr, ins_bytes, api_function):
        if ins_bytes.endswith(b"\x0a"):
            if api_function.startswith("<dnfile.mdtable.MemberRefRow"):
                return
            self.disassembly.addr_to_api[from_addr] = api_function

    def analyzeFunction(self, pe, start_addr, method_body):
        LOGGER.debug("analyzeFunction() starting analysis of candidate @0x%08x", start_addr)
        state = FunctionAnalysisState(start_addr, method_body.instructions[0].offset, self.disassembly)
        for insn in method_body.instructions:
            state.setNextInstructionReachable(True)
            i_bytes = insn.get_bytes()
            i_address = insn.offset
            i_size = len(i_bytes)
            i_mnemonic = str(insn.opcode)
            i_op_str = format_operand(pe, insn.operand)
            # debug output for all instructions
            if False:
                from smda.cil.CilInstructionEscaper import CilInstructionEscaper
                from smda.common.SmdaInstruction import SmdaInstruction
                smda_ins = SmdaInstruction([i_address, i_bytes.hex(), i_mnemonic, i_op_str])
                escaped = CilInstructionEscaper.escapeBinary(smda_ins)
                print(
                    f"{escaped:<20}"
                    + "{:04X}".format(insn.offset)
                    + "    "
                    + f"{' '.join('{:02x}'.format(b) for b in insn.get_bytes()) : <20}"
                    + f"{str(insn.opcode) : <15}"
                    + format_operand(pe, insn.operand)
                )
            # https://en.wikipedia.org/wiki/List_of_CIL_instructions
            if i_mnemonic in ["ret"]:
                state.setNextInstructionReachable(False)
            if i_mnemonic in [
                    "beq", "beq.s", 
                    "bge", "bge.s", "bge.un", "bge.un.s",
                    "bgt", "bgt.s", "bgt.un", "bgt.un.s",
                    "ble", "ble.s", "ble.un", "ble.un.s",
                    "blt", "blt.s", "blt.un", "blt.un.s",
                    "bne.un", "bne.un.s", 
                    "br", "br.s", 
                    "brfalse", "brfalse.s", 
                    "brinst", "brinst.s", 
                    "brnull", "brnull.s", 
                    "brtrue", "brtrue.s", 
                    "brzero", "brzero.s"]:
                target = int(i_op_str, 16)
                state.addCodeRef(i_address, target, by_jump=True)
            if i_mnemonic in ["jmp"]:
                raise Exception("Found unhandled CIL jmp instruction, report back its structure and have Daniel fix it.")
                target = int(i_op_str, 16)
                state.addCodeRef(i_address, target, by_jump=True)
                state.setNextInstructionReachable(False)
            if i_mnemonic in ["ldstr"]:
                # we possibly want to extract and collect these and put them in the stringref part of SmdaFunction
                self.disassembly.addStringRef(start_addr, i_address, i_op_str[1:-1])
            if i_mnemonic in ["call", "callvirt"]:
                self._updateApiInformation(i_address, i_bytes, i_op_str)
                # https://blog.objektkultur.de/about-tail-recursion-in-.net/
                if state.prev_opcode.startswith("tail"):
                    state.setNextInstructionReachable(False)
                if i_bytes.endswith(b"\x06"):
                    operand = resolve_token(pe, insn.operand)
                    if isinstance(operand, dnfile.mdtable.MethodDefRow):
                        # override operand string with "address" of the method
                        method_name = self.cil_label_provider.decodeSymbolName(operand.Name.value)
                        i_op = f"0x{self.cil_label_provider.getAddress(method_name):x}"
            if i_mnemonic in ["throw"]:
                state.setNextInstructionReachable(False)
            if i_mnemonic in ["switch"]:
                next_addrs = []
                for target in insn.operand:
                    next_addrs.append(target)
                    state.addCodeRef(i_address, target, by_jump=True)
            state.prev_opcode = i_mnemonic
            state.addInstruction(i_address, i_size, i_mnemonic, i_op_str, i_bytes)
        state.label = self.resolveSymbol(state.start_addr)
        analysis_result = state.finalizeAnalysis()
        return state

    def analyzeBuffer(self, binary_info, cbAnalysisTimeout=None):
        LOGGER.debug("Analyzing buffer with %d bytes @0x%08x", binary_info.binary_size, binary_info.base_addr)
        self._updateLabelProviders(binary_info)
        self.disassembly = DisassemblyResult()
        self.disassembly.smda_version = self.config.VERSION
        self.disassembly.setBinaryInfo(binary_info)
        self.disassembly.binary_info.architecture = "cil"
        self.disassembly.analysis_start_ts = datetime.datetime.now(datetime.timezone.utc)
        self.disassembly.language = "cil"

        LOGGER.debug("Starting parser-based analysis.")
        pe = dnfile.dnPE(data=binary_info.raw_data)
        all_instruction_offsets = set([])
        for row in pe.net.mdtables.MethodDef:
            if not row.ImplFlags.miIL or any((row.Flags.mdAbstract, row.Flags.mdPinvokeImpl)):
                # skip methods that do not have a method body
                continue
            try:
                method_body = CilMethodBody(DnfileMethodBodyReader(pe, row))
            except MethodBodyFormatError as e:
                LOGGER.error(e)
                continue
            if not method_body.instructions:
                continue
            if cbAnalysisTimeout and cbAnalysisTimeout():
                break
            function_result_state = self.analyzeFunction(pe, method_body.offset, method_body)
        # package up and finish
        self.disassembly.analysis_end_ts = datetime.datetime.now(datetime.timezone.utc)
        if cbAnalysisTimeout():
            self.disassembly.analysis_timeout = True
        return self.disassembly



#!/usr/bin/env python3
import logging
import re
import struct
import codecs

def occurrences(string, sub):
    # https://stackoverflow.com/a/2970542
    count = start = 0
    while True:
        start = string.find(sub, start) + 1
        if start > 0:
            count += 1
        else:
            return count

LOGGER = logging.getLogger(__name__)


class CilInstructionEscaper:
    """ 
    Escaper to abstract information from disassembled instructions. 
    Based on https://en.wikipedia.org/wiki/List_of_CIL_instructions
    """

    _aritlog_group = [
        "add", "add.ovf", "add.ovf.un", "and", "div", "div.un", "mul", "mul.ovf", 
        "mul.ovf.un", "neg", "not", "or", "rem", "rem.un", "shl", "shr", "shr.un", "xor"
    ]
    
    _cfg_group = [
        # branching
        "beq", "beq.s", "bge", "bge.s", "bge.un", "bge.un.s", "bgt", "bgt.s", "bgt.un", 
        "bgt.un.s", "ble", "ble.s", "ble.un", "ble.un.s", "blt", "blt.s", "blt.un", 
        "blt.un.s", "bne.un", "bne.un.s", "br", "br.s", "brfalse", "brfalse.s", 
        "brinst", "brinst.s", "brnull", "brnull.s", "brtrue", "brtrue.s", "brzero", 
        "brzero.s", "jmp", "ret", "switch",
        # Exception handling control flow
        "leave", "leave.s", "throw", "rethrow", "endfilter", "endfault", "endfinally",
        # Calls
        "call", "calli", "callvirt", 
        # Comparisons
        "ceq", "cgt", "cgt.un", "clt", "clt.un", "ckfinite"
    ]
    
    _mem_group = [
        "ldarga", 
        "ldarga.s", "starg", "starg.s", 
        "stelem", "stelem.i", "stelem.i1", "stelem.i2", "stelem.i4", "stelem.i8", 
        "stelem.ref", 
        "stind.i", "stind.i1", "stind.i2", "stind.i4", "stind.i8", 
        "stind.ref", 
        "stfld", "stobj", "stsfld",
        # copy
        "cpblk", "initblk", "initobj", "localloc", "ldtoken",
        # special 
        "arglist"
    ]
    
    _stack_group = [
        "dup", "pop",
        "ldarg", "ldarg.0", "ldarg.1", "ldarg.2", "ldarg.3", "ldarg.s", 
        "ldelem", "ldelem.i", "ldelem.i1", "ldelem.i2", "ldelem.i4", 
        "ldelem.i8", "ldelem.ref", "ldelem.u1", "ldelem.u2", 
        "ldelem.u4", "ldelem.u8", "ldelema", "ldfld", "ldflda", "ldlen", "ldobj", 
        "ldsfld", "ldsflda", "ldstr", 
        # Stack operations moved from mem_group
        "ldc.i4", "ldc.i4.0", "ldc.i4.1", "ldc.i4.2", "ldc.i4.3", "ldc.i4.4",
        "ldc.i4.5", "ldc.i4.6", "ldc.i4.7", "ldc.i4.8", "ldc.i4.m1", "ldc.i4.M1", 
        "ldc.i4.s", "ldc.i8", 
        "ldind.i", "ldind.i1", "ldind.i2", "ldind.i4", "ldind.i8", 
        "ldind.ref", "ldind.u1", "ldind.u2", "ldind.u4", "ldind.u8",
        "ldloc", "ldloc.0", "ldloc.1", "ldloc.2", "ldloc.3", "ldloc.s", "ldloca", 
        "ldloca.s", "ldnull",
        "stloc", "stloc.0", "stloc.1", "stloc.2", "stloc.3", "stloc.s",
        # Box/Unbox operations
        "box", "unbox", "unbox.any",
        # Function pointer to stack
        "ldftn", "ldvirtftn",
        # Conversion operations (all work with stack)
        "conv.i", "conv.i1", "conv.i2", "conv.i4", "conv.i8",
        "conv.ovf.i", "conv.ovf.i1", "conv.ovf.i2", "conv.ovf.i4", "conv.ovf.i8",
        "conv.ovf.u", "conv.ovf.u1", "conv.ovf.u2", "conv.ovf.u4", "conv.ovf.u8",
        "conv.r.un", 
        "conv.u", "conv.u1", "conv.u2", "conv.u4", "conv.u8"
    ]
    
    _prefix_group = [
        "constrained.", "readonly.", "unaligned.", "volatile.",
        "no.typecheck", "no.rangecheck", "no.nullcheck", "tail",
    ]

    _privileged_group = [
        "break"
    ]

    _crypto_group = [
        ]

    _float_group = [
        # while also stack instructions, float identity is more important for characteristics
        "stelem.r4", "stelem.r8", 
        "stind.r4", "stind.r8",
        "ldelem.r4", "ldelem.r8", 
        "ldc.r4", "ldc.r8",
        "ldind.r4", "ldind.r8",
        "conv.r4", "conv.r8",
    ]

    _nop_group = [
        "nop"
    ]

    _xmm_group = [
    ]
    _vm_group = [
    ]

    _registers = [
    ]
    _segment_registers = [
    ]
    _extended_registers = [
    ]
    _control_registers = [
    ]

    @staticmethod
    def escapeMnemonic(mnemonic):
        mnemonic = mnemonic.split(" ")[-1]
        if mnemonic in CilInstructionEscaper._aritlog_group:
            return "A"
        elif mnemonic in CilInstructionEscaper._cfg_group:
            return "C"
        elif mnemonic in CilInstructionEscaper._mem_group:
            return "M"
        elif mnemonic in CilInstructionEscaper._stack_group:
            return "S"
        elif mnemonic in CilInstructionEscaper._privileged_group:
            return "P"
        elif mnemonic in CilInstructionEscaper._crypto_group:
            return "Y"
        elif mnemonic in CilInstructionEscaper._float_group:
            return "F"
        elif mnemonic in CilInstructionEscaper._xmm_group:
            return "X"
        elif mnemonic in CilInstructionEscaper._vm_group:
            return "V"
        elif mnemonic in CilInstructionEscaper._nop_group:
            return "N"
        elif mnemonic == "error":
            return "U"
        else:
            LOGGER.error("********************************************** Unhandled mnemonic: %s", mnemonic)
            return "U"
        return mnemonic

    @staticmethod
    def escapeField(op_field, escape_registers=True, escape_pointers=True, escape_constants=True):
        op_field = op_field.strip()
        escaped_field = ""
        if op_field == "":
            return ""
        if escape_registers:
            # there are no registers in CIL, it is a stack-based machine
            pass
        if escape_pointers:
            # if we do not have an immediate, we assume it is a pointer
            if not (op_field.startswith("-")
                    or op_field.startswith("0")
                    ):
                escaped_field = "PTR"
        if escape_constants:
            try:
                op_as_int = int(op_field)
                escaped_field = "CONST"
            except:
                pass
            try:
                op_as_int = int(op_field, 16)
                escaped_field = "CONST"
            except:
                pass
            if ":" in op_field:
                escaped_field = "CONST"
        if not escaped_field:
            escaped_field = op_field
        return escaped_field

    @staticmethod
    def escapeOperands(ins, offsets_only=False):
        opstring = ins.operands
        op_fields = opstring.split(",")
        esc_regs = True
        esc_consts = True
        if offsets_only:
            if ins.mnemonic in CilInstructionEscaper._cfg_group and ins.mnemonic not in [
                    "ceq", "cgt", "cgt.un", "clt", "clt.un", "ckfinite"]:
                return "OFFSET"
            esc_regs = False
            esc_consts = False
        escaped_fields = []
        for op_field in op_fields:
            escaped_fields.append(CilInstructionEscaper.escapeField(op_field, escape_registers=esc_regs, escape_constants=esc_consts))
        return ", ".join(escaped_fields)

    @staticmethod
    def escapeToOpcodeOnly(ins):
        opcode_length = 2 if ins.bytes.startswith("fe") else 1
        return ins.bytes[:opcode_length*2] + "?" * (len(ins.bytes) - opcode_length*2)

    @staticmethod
    def escapeBinary(ins, escape_intraprocedural_jumps=False, lower_addr=None, upper_addr=None):
        escaped_sequence = ins.bytes
        if ins.mnemonic in CilInstructionEscaper._cfg_group and ins.mnemonic not in ["ceq", "cgt", "cgt.un", "clt", "clt.un", "ckfinite"]:
            escaped_sequence = CilInstructionEscaper.escapeBinaryJumpCall(ins, escape_intraprocedural_jumps)
            return escaped_sequence
        if ins.mnemonic in ["newobj", "ldobj", "stobj", "newarr", "stsfld", "stfld", "ldsfld", "ldfld", "ldflda", "box", "unbox", "unbox.any", "isinst", "castclass", "ldtoken"]:
            escaped_sequence = ins.bytes[:2] + "?" * (len(ins.bytes) - 4) + ins.bytes[-2:]
            return escaped_sequence
        if ins.mnemonic in ["initobj", "ldftn", "constrained."]:
            escaped_sequence = ins.bytes[:4] + "?" * (len(ins.bytes) - 6) + ins.bytes[-2:]
            return escaped_sequence
        if False:
            # 20250211 - we don't know yet how allocations look like in CIL, so we leave this out for now
            escaped_sequence = CilInstructionEscaper.escapeBinaryPtrRef(ins)
        if ins.operands.startswith("0x") or ins.operands.startswith("-"):
            # 20250211 - when comparing with Intel, we onl want to wildcard addresses within our range - should we not wildcard at all here?
            pass
            # escaped_sequence = CilInstructionEscaper.escapeBinaryValue(ins, escaped_sequence, None)
        return escaped_sequence

    @staticmethod
    def escapeBinaryJumpCall(ins, escape_intraprocedural_jumps=False):
        escaped_sequence = ins.bytes
        if ins.mnemonic in CilInstructionEscaper._cfg_group and ins.mnemonic not in ["ceq", "cgt", "cgt.un", "clt", "clt.un", "ckfinite"]:
            if escape_intraprocedural_jumps and ins.mnemonic in [
                    "beq", "beq.s", "bge", "bge.s", "bge.un", "bge.un.s", "bgt", "bgt.s", "bgt.un", 
                    "bgt.un.s", "ble", "ble.s", "ble.un", "ble.un.s", "blt", "blt.s", "blt.un", 
                    "blt.un.s", "bne.un", "bne.un.s", "br", "br.s", "brfalse", "brfalse.s", 
                    "brinst", "brinst.s", "brnull", "brnull.s", "brtrue", "brtrue.s", "brzero", "switch"]:
                escaped_sequence = CilInstructionEscaper.escapeToOpcodeOnly(ins)
            elif ins.mnemonic in ["jmp", "call", "calli", "callvirt"]:
                escaped_sequence = ins.bytes[:2] + "?" * (len(ins.bytes) - 4) + ins.bytes[-2:]
        return escaped_sequence

    @staticmethod
    def escapeBinaryPtrRef(ins):
        escaped_sequence = ins.bytes
        return escaped_sequence

    @staticmethod
    def escapeBinaryValue(ins, escaped_sequence, value):
        return CilInstructionEscaper.escapeToOpcodeOnly(ins)

    @staticmethod
    def getByteWithoutPrefixes(ins_bytes):
        # if I understand correctly, there are only prefix instructions and not prefix bytes within instructions
        return ins_bytes
